//
//    MTLineMon searches for line features in the spectrum of a channel using
//    multitaper spectral techniques. 
//    The configuration for LineMon is specified by an ascii file 
//    passed to LineMon() as an argument. The file contains a single
//    line:
//
//    Line 1: <channel> <tAverage> 
//
//    The fields have the following meaning:
//    <channel>      The name of the channel from which the data are derived.
//    <tAverage>     The averaging time over which the spectrum is taken.
//    
//
//    In addition, there are a number of command line switches:
//         MTLineMon -conf <config>         : name of configuration file 
//                  [-adaptive or -hires]   : high resolution or adaptive multitaper spectrum
//                  [-nWin <number of Windows: default 5>] : number of tapers used 
//                  [-nwdt <NWdt: default 3.0>]  : NWdt 
//                  [-paddingfactor <default: 4> : Factor by which to pad the time series
//                  [-freqRange <default: 0 8192>]  : Frequency range in which to search for and remove lines
//                  [-debug] : produce more verbose output
//                  [-nofinalspec] : don't calculate spectrum with lines removed 
//                  [-log] : write log files containing information on lines found
//                  [-maxSegments <default -1 (no limit)>] : maximum number of data segemnts of length tAverage
//                                                           to process         
//
#include <cstdlib>
#include <math.h>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <unistd.h>  
#include "MTLineMon.hh"

//-->  The next three lines are needed if you are going to generate triggers.
//     The descriptive title in PIDTITLE should the monitor function.
#define PIDCVSHDR "$Header: /home/ottewill/myDMT/MTLineMon/MTLineMon.cc"
#define PIDTITLE  "Multitaper Line Monitor"
#include "ProcIdent.hh"

EXECDAT(MTLineMon);

using namespace std;

MTLineMon::MTLineMon(int argc, const char* argv[]) 
  :DatEnv(argc, argv), MonServer("MTLineMon"), 
  liveTime(0.0), maxLines(100), maxPasses(500), mSpectralType(HIRES),
  nWin(5),nwdt(3.0),paddingFactor(4),lowFreq(0.0),highFreq(0.0),
  mDump(0),nDump(2),maxDump(0),mMaxSegment(-1),nBuffer(8192),
  mGetFinalSpec(true),mWriteLogFiles(false),mProduceTrends(false),mSnapshot(false), 
  mOSC(getDacc())
{
   //set defaults 
   setDebug(0);
   sprintf(lockStatus,"%s","");


   // Parse command line
   const char* config="";
   for (int i=1 ; i<argc ; i++) {
     std::string argi = argv[i];
     if (argi == "-conf" || argi == "-c") {
       config = argv[++i];
     } else if (argi == "-adaptive") {
       mSpectralType=ADAPTIVE;
     }  else if (argi == "-hires") { // redundant but in for completeness 
       mSpectralType=HIRES;
     }  else if (argi == "-nWin" || argi == "-nwin" ) {
       nWin = atoi(argv[++i]);
     } else if (argi == "-nwdt" || argi == "-NWdt" ) {
       nwdt = atof(argv[++i]);
     }  else if (argi == "-paddingfactor" || argi == "-paddingFactor" ) {
       paddingFactor = atoi(argv[++i]);
     }  else if (argi == "-freqRange" || argi == "-Range"  || argi == "-freqrange" || argi == "-range" ) {
       lowFreq = atof(argv[++i]);
       highFreq = atof(argv[++i]);
     }  else if (argi == "-debug" || argi == "-d") {
       setDebug(1);
     } else if (argi == "-nofinalspec" || argi == "-noFinalSpec") {
       mGetFinalSpec = false;
     } else if (argi == "-log") {
       mWriteLogFiles=true;
     } else if (argi == "-maxSegments") {
       mMaxSegment = atoi(argv[++i]);
     } else if (argi == "-help") {
        usage();
        exit(0);
	return;
     } else {
            cerr << "Invalid argument: " << argi << endl;
	    usage();
            exit(INVALID_ARG);
	    return;
     }         
   }

   readConfigFile(config);

   // get OUTPUT directories
   const char *dmthtmlout= getenv("DMTHTMLOUT");
   if (dmthtmlout) htmlDir = string(dmthtmlout)+"/";
   if (getDebug()) cerr << "MTLineMon: HTML directory " << htmlDir << endl;

   // Note add check if directory for output directories exist
     
   // set html file name
   htmlFile=htmlDir+"MT"+string(chanv.mName)+".html";
   if (getDebug()) cerr <<"MTLineMon: output html file name "<< htmlFile << endl;


   //----------------------------------  Define and initialize parameters.
   mDict.addPar("HTMLFile",   htmlFile);
   mDict.addPar("Stride",     sampleTime);
   mDict.addPar("TrigEnable", 0);
   mDict.addPar("MonName",    string(argv[0]));
   mDict.addPar("OSCFile",    string(""));
   mDict.addPar("TrendCount", 1);
   mDict.addPar("TrendFile",  string(""));              
   
   // set server Name
   
   string serverName = mDict.getString("MonName")+string(chanv.mName);
   setServerName(serverName.c_str());
   
   chanv.mTS = new TSeries(Time(0), Interval(sampleTime));
   // chanv.mTS = new TSeries();
   getDacc().addChannel(chanv.mName, 0, &(chanv.mTS));  //  Set channel name
   getDacc().setBuffer(0);       // Set to take up data as requested
   getDacc().setStride(Interval(mDict.getDouble("Stride")));   // No of seconds of data to take up
    

   initialTS = new TSeries(Time(0),Interval(sampleTime),nBuffer);
   finalTS = new TSeries(Time(0),Interval(sampleTime),nBuffer);


   // read OperStateCondition file
   FILE *fp;
   if(strlen(lockStatus)) {
     if( (fp=fopen("/export/home/ops/pars/LockLoss.conf","r"))==NULL ) {
       cerr << "Error: can't open /export/home/ops/pars/LockLoss.conf "<< endl;
       fclose(fp);
       if( (fp=fopen("LockLoss.conf","r"))==NULL ) {
         cerr << "Error:         can't open LockLoss.conf file"<< endl;
         cerr << "copy LockLoss.conf file into local directory"<< endl;
         exit(NO_LOCKLOSS);
       }
       else{
         fclose(fp);
         mDict.addPar("OSCFile",string("LockLoss.conf"));
       }
     }
     else{
       fclose(fp);
       mDict.addPar("OSCFile",string("/export/home/ops/pars/LockLoss.conf"));
     }
   }
     
   if (!mDict.getString("OSCFile").empty()) {
     cout<<"read lock loss file: "<<mDict.getString("OSCFile").c_str()<<endl;
     mOSC.readConfig(mDict.getString("OSCFile").c_str());
   }

   if (mProduceTrends) {
     
     cerr << " ACO : trends to be implemented " << endl;
     
     const char *dmttrendout= getenv("DMTTRENDOUT");
     if (dmttrendout) trendDir = string(dmttrendout)+"/";
     if (getDebug()) cerr << "MTLineMon: Trend directory " << trendDir << endl;


     const char *dmtoutput= getenv("DMTOUTPUT");
     if (dmtoutput) outputDir = string(dmtoutput)+"/";
     if (getDebug()) cerr << "MTLineMon: Index directory " << outputDir << endl;

   //    mTrend.setName(serverName.c_str());  
//       mTrend.setType(Trend::kMinute);
//       mTrend.setFrameCount(1);   
//       if (chanv.mName && chanv.mName[2] == ':') {
//         string ifo(string(chanv.mName), 2);
//         mTrend.setIFO(ifo.c_str());
//       }       
 
   }      
    
   mSegment=0; 

}
   
void
MTLineMon::ProcessData() {    
  if (!mSegment) {  // Is this the first data segment.  If so we determine the sample rate and 
                      // allocate storage on the first pass

    startTime=(chanv.mTS->getStartTime()).getS();
    if (getDebug()) {
      char UTCdate[256];
      TimeStr(chanv.mTS->getStartTime(), UTCdate,"%H:%N:%S %M %d %Y %Z");
      std::cerr << "# MTLineMon: starting at time " << UTCdate << std::endl;
    }



  
    chanv.mFS = new FSeries(*(chanv.mTS));
          MTSpecInit = new FSpectrum(*(chanv.mFS));
    if (mGetFinalSpec) MTSpecFinal = new FSpectrum(*(chanv.mFS));
    SpecInit = new FSpectrum(*(chanv.mFS));
    SpecFinal = new FSpectrum(*(chanv.mFS));


    chanv.mRate=1./(chanv.mTS->getTStep());  // sample rate
    chanv.mPoint=int(sampleTime*chanv.mRate);
    if (getDebug()) std::cerr << "# MTLineMon: Channel: " << chanv.mName << "\t No of points:"
				<< chanv.mPoint << std::endl;	
 
    paddedLength=paddingFactor*chanv.mPoint;   // number of points with zero pading
    nFreq=1+paddedLength/2;     // number of frequencies bins including DC, Nyquist 
    if (nBuffer > chanv.mPoint) nBuffer=chanv.mPoint;
    if (getDebug()) cerr << "# CorrMon: dmtviewer TSeries buffer size " << nBuffer << endl;


    decimationFactor=chanv.mPoint/nBuffer;
    if (mDebug) cerr << "# CorrMon: dmtviewer TSeries decimation " << decimationFactor << endl;


   
    if (getDebug()) std::cerr << "# MTLineMon:   Padded Length: " << paddedLength 
				<< "\t Number of Frequency bins: " << nFreq << std::endl;
    // allocate arrays
    lineList=vector<RemovedLines> (maxLines);
    lineCount=vector<unsigned long> (nFreq);


    // The critical value of F, below which we do not consider the line component significant 
    // See Percival and Walden Sec 10.11 and pg 513 (top) for discussion of critical value 
    fCritical=fCrit();
    if (getDebug()) std::cerr << "Critical F-value: " << fCritical << endl;

    // compute number of bins over which F-test is correlated 
    wwidth=int(nwdt*paddingFactor);
    nearNyquist=wwidth;

    if (highFreq==0.0) highFreq=chanv.mRate;
    lowBin=int(floor(lowFreq*paddingFactor*sampleTime/2));
    highBin=int(ceil(highFreq*paddingFactor*sampleTime/2));
    if (getDebug()) std::cerr << "Bin range: " << lowBin << " to " << highBin << endl;


    // serve the data to the DMT viewer
    string title;
    title=string(chanv.mName)+": Initial Multitaper Spectrum";
    serveData(title.c_str(), MTSpecInit, "Initial Multitaper Spectrum");
    title=string(chanv.mName)+": Final Multitaper Spectrum";
    serveData(title.c_str(), MTSpecFinal,"Initial Multitaper Spectrum");  
    title=string(chanv.mName)+": Initial Spectrum";   
    serveData(title.c_str(), SpecInit, "Initial Spectrum");
    title=string(chanv.mName)+": Final Spectrum";   
    serveData(title.c_str(), SpecFinal,"Final Spectrum");
    title=string(chanv.mName)+": Initial Time Series";   
    serveData(title.c_str(), initialTS);
    title=string(chanv.mName)+": Final Time Series (Lines Removed)";   
    serveData("Final Time Series", finalTS,"(Lines Removed)");
  }

  if((mDump == 0 && maxDump == 0) || maxDump > 100000){
    if(maxDump == 0){
      char w[256];
      sprintf(w,"%ld",long((chanv.mTS->getStartTime()).totalS()));
      indexFile=trendDir+string(chanv.mName)+"-"+string(w)+".txt";
      // while((w = strchr(indexFile,':'))) w[0] = '-';
      if (getDebug()) cout <<"#MTLineMon: index file name "<< indexFile << endl;
    }        
    maxDump = 0;
  }    
   
  if (chanv.mTS->isGood() == false) {
    cerr << "MTLineMon: Data not valid" << endl;
    exit(BAD_DATA);
    return;
  }

//    if(strlen(lockStatus) && !(mOSC.satisfied(lockStatus))) {
//      if (getDebug()) cerr <<"#MTLineMon: Out of Lock" << endl; 
//    }

  // subtract the mean value from the data set 
  double mean=chanv.mTS->getAverage();
  chanv.mTS->Convert(DVector::t_float);
  *(chanv.mTS)-=mean;

  currentTime = double((chanv.mTS->getStartTime()).getS());
  liveTime += sampleTime;

  // save initial data set (with mean removed)
  *initialTS=TSeries(chanv.mTS->decimate(decimationFactor));
  chanv.mFS->setData(*(chanv.mTS));
  *SpecInit=FSpectrum(*(chanv.mFS));
  
  // remove the spectral lines from the data set
  removeSpectralLines(0,nFreq);   
  
  *finalTS=TSeries(chanv.mTS->decimate(decimationFactor));
  chanv.mFS->setData(*(chanv.mTS));
  *SpecFinal=FSpectrum(*(chanv.mFS));

  // print out a list of lines found
  if (getDebug()) {
    char UTCdate[256];
    TimeStr(chanv.mTS->getStartTime(), UTCdate,"%H:%N:%S %M %d %Y %Z");
    std::cerr << "Section starting at time " << UTCdate << std::endl;
    std::cerr << "Total number of lines found: " << nRemoved << std::endl;
    for (int i=0;i<nRemoved;i++) {
      std::cerr << "Found line frequency " << 0.5*lineList[i].index*chanv.mRate/nFreq 
		<< "Hz amplitude " <<  2.0*lineList[i].FC.Mag()  
		<< " phase " << lineList[i].FC.Arg() 
		<< " (F-test " << lineList[i].fvalue  << ")" << std::endl;
    }
  }

  if (mWriteLogFiles) {
    char* dmtout = getenv("DMTOUTPUT");
    string fmtstr;
    if (dmtout == NULL) 
      fmtstr = std::string("");  
    else
      fmtstr = std::string(dmtout)+"/";
    fmtstr+=string("MTLineMon%s.log");
    char fileName[128];
    TimeStr(chanv.mTS->getStartTime(), fileName,fmtstr.c_str());
    std::ofstream ostr(fileName);
    ostr.precision(4);
    ostr.setf(ios::fixed,ios::floatfield);
    if (ostr.good()) {
      for (int i=0;i<nRemoved;i++) {   
	ostr << setw(10) << 0.5*lineList[i].index*chanv.mRate/nFreq << "\t"
	     << setw(10) << 2.0*lineList[i].FC.Mag()  << "\t" 
	     << setw(10) << lineList[i].FC.Arg() << "\t" 
	     << setw(10) << lineList[i].fvalue << "\t" 
	     << setw(6) << lineList[i].dof << std::endl;
      }
    } else {
      std::cerr << "#MTLineMon: Error opening file " << fileName << std::endl;
      finish();
    }
    
  }

  mDump++;

  if(mDump == nDump){
    if (!maxDump) appendMode = 0;
    if (getDebug()) cerr << "MTLineMon: Write html now" << endl;
    DumpHTML(chanv.mName, htmlFile.c_str());  
    appendMode = 1;
    maxDump += mDump;
    mDump = 0;
  }

  mSegment++;
  if (mSegment == mMaxSegment) finish();  //  if a maximum number of passes is specified
}


MTLineMon::~MTLineMon() {
  //---------------------------------  Clean up at the end
  getDacc().rmChannel(chanv.mName);
  delete initialTS;

  //  mTrend.writeIndex(indexFile);
  mTrend.close();
                               

  if (getDebug()) std::cerr << "MTLineMon finished" << std::endl;
}               


void MTLineMon::Attention(void) {
    if (getDebug()) std::cerr << "MTLineMon: Attention Interrupt" << std::endl;
    MonServer::Attention();
}


//   This is the critical value of f for npoints number of points in a
//   data-set and nWin tapers, according to Thompson's rule of thumb
//   (Percival & Walden paragraph at top of pg 513) that the confidence
//   should exceed 1-1/n, combined with the explicit formula for
//   the appropriate percentage point given on pg 501.
double MTLineMon::fCrit() 
{
	double nm1=double(nWin-1);
	return nm1*(pow(double(chanv.mPoint),1.0/nm1)-1.0);
}

//***************************************************************

// Checks if the argument is a power of 2 less than 2^32 
bool MTLineMon::isPowerOfTwo(const int test) {
        int k=1;
	for (int i=0;i<32;i++) {
		if (k==test) return true;
		k+=k;
	}
	return false;
}


// A routine to check if there are any other nearby points arising from the same 
// spectral line - if so we only want to remove the most significant one   
bool MTLineMon::emptyWindow(fComplex *workSpace,int index,int wwidth) {
  // step through the +- wwidth 
  for (int i=-wwidth;i<=wwidth;i++) {
    // if the frequency bin is below 0 or above the end, set correctly 
    int bin=index+i;
    if (bin<0) bin=0;
    if (bin>=paddedLength/2) bin=paddedLength/2-1;
    // test if there are any non-zero entries in workspace 
    if (workSpace[bin].Real() != 0.0 || workSpace[bin].Imag() != 0.0  ) return false;
  }
  return true;
}

// this routine sorts into increasing frequency order and combines any duplicates 
void MTLineMon::sortLines() {
  // Sort into increasing frequency index order 
  // hack for sorting by index
  // for (int i=nRemoved;i<maxLines;i++) lineList[i].index=paddedLength/2; 
  sort(lineList.begin(),lineList.begin()+nRemoved,compareIndex()); 

  // search/combine duplicates 
  int bot=0;
  int top=0;
  int didRemove=0;
  while (top<nRemoved){
    bot=top+1;
    while (bot<nRemoved && lineList[bot].index==lineList[top].index) {
      lineList[top].FC+=lineList[bot].FC;
      lineList[bot++].fvalue=0.0;
      didRemove++;
    }
    top=bot;
  }

  nRemoved-=didRemove;


  // sort by significance
  sort(lineList.begin(),lineList.begin()+nRemoved,compareFValue()); 

  // sort the most significant by frequency
  sort(lineList.begin(),lineList.begin()+nRemoved,compareIndex());

  return;
}


//***************************************************************
//***************************************************************
void MTLineMon::removeSpectralLines(const int lowBin,const int highBin) {

  FSpectrum *MTSpec=new FSpectrum();

  // check that the padded array has a length = integer power of 2 
  if (!isPowerOfTwo(paddedLength)) {
    std::cerr<< "removeSpectralLines() Error: the paddedLength must"
	     << " be an integer power of 2 <= 2^31, but is " 
	     << paddedLength << std::endl;
    abort();
  }

  // The number of degrees of freedom (only useful for adaptive, not hires tapered spectra)
  float *dof=new float[nFreq];
  // Array of values of the F-test 
  float *fvalues=new float[nFreq];
  // Array to hold the associated Fourier coefficient
  fComplex *cEstimate=new fComplex[nFreq];
  // Array that will be used to do IFFT to remove the lines 
  FSeries *workFSeries=new FSeries;
  fComplex *workSpace=new fComplex[paddedLength/2];
  // Working structure to hold frequencies and f-values  
  vector<RemovedLines> workList(nFreq);
  if (dof==NULL || fvalues==NULL || cEstimate==NULL) {
    std::cerr << "# MTLineMon::removeSpectralLines() : Insufficient memory" << std::endl;
    abort();
  }

  nRemoved=0;
  bool didRemove=false;
  int nLines=maxLines; // dummy value to ensure we pass through loop the first time
  int passes=0;

  // run through loop removing lines, ending when no additional lines removed 
  while (passes<wwidth && nLines!=0 && passes<maxPasses) {  

    // subtract the mean value from the data set 
    *(chanv.mTS)-=chanv.mTS->getAverage();
  
     // calculate the multitaper spectrum 
    MTSpectrum(4 /* normalization choice */,float(1.0) /* sample rate */,
               MTSpec,dof,fvalues,cEstimate);
  
    // if first pass through store spectrum in MTSpecInit otherwise in MTSpecFinal  
    if (passes==0) {
      *MTSpecInit=FSpectrum(*MTSpec);
    }

    passes++;
 
    // (skip DC, Nyquist) looking for frequency bins which pass the F-test 
    nLines=0;
    for (int i=1;i<nFreq-nearNyquist-1;i++) {
      if (fvalues[i]>fCritical){
	workList[nLines].index=i;
	workList[nLines].fvalue=fvalues[i];
	workList[nLines].dof=dof[i];
	workList[nLines].FC=cEstimate[i];
	// pass back estimates of the derivative (f+ - f-)/2 
	workList[nLines].dFC=0.5*(cEstimate[i+1]-cEstimate[i-1]);
	// pass back estimates of the second derivative (f+ - 2f + f-) 
	workList[nLines].d2FC=cEstimate[i+1]-2.0*cEstimate[i]+cEstimate[i-1];
	nLines++;	
      }		  
    }
    if (nLines==0) break;


    for (int i=nLines;i<nFreq;i++) workList[nLines].fvalue=0.0;

    // Now sort into order of decreasing significance to ensure we keep
    // the most significant line when we discard multiple detections
    sort(workList.begin(),workList.begin()+nLines,compareFValue()); 

    didRemove=false;
  
    for (int i=0;i<paddedLength/2;i++) workSpace[i]=fComplex(0.0,0.0);
    for (int i=0;i<nLines && nRemoved<maxLines;i++) {
      // check to see if we are removing any lines within wwidth bins 
      if (emptyWindow(workSpace,workList[i].index,wwidth)) {
	// if there are no nearby lines already removed, add lines to remove list 
	lineList[nRemoved]=workList[i];
	workSpace[workList[i].index]=-2.0*(~lineList[nRemoved].FC);	    
	nRemoved++;
	didRemove=true;
      }			
    }

    // efficiently calculate the sum of the spectral lines to remove 
    // the conventions are almost exactly Percival and Walden example eqns 
    // 20 lines down on page 513
    if (didRemove) {
      workFSeries=new FSeries(0.0,
			      1/(chanv.mTS->getInterval()*double(paddingFactor)),
			      chanv.mTS->getStartTime(),
			      double(paddingFactor)*chanv.mTS->getInterval(),
			      size_t(paddedLength/2),workSpace);
      TSeries *t1=new TSeries(*workFSeries); 
      delete workFSeries;
      TSeries t2=t1->extract(chanv.mTS->getStartTime(),chanv.mTS->getInterval());
      delete t1;
      t2*=sqrt(sampleTime*paddingFactor/2.0);  // ACO: NR/DMT convention issue
      *(chanv.mTS)+=t2;
    }
  }
 
  // If last pass through loop changed data, need to find spectrum. 
  // This will only happen if the number of lines removed is maxlines.  
  // However it may be that the we remove maxlines but didRemove==false.
  if (mGetFinalSpec) {
    // If we removed no points, final spectrum = initial spectrum 
    if (nRemoved==0)  {
      *MTSpecFinal=FSpectrum(*MTSpecInit);
    } else {
      MTSpectrum(4 /* normalization choice */,float(1.0) /* sample rate */,
	       MTSpec,dof,fvalues,cEstimate);
      *MTSpecFinal=FSpectrum(*MTSpec);
    }
  }

  // combine any duplicates and sort into increasing frequency bin order 
  sortLines();


  for (int i=0;i<nRemoved;i++) {
    lineCount[lineList[i].index]++;
  }

  // de-allocate the storage space 
  delete [] dof;
  delete [] fvalues;
  delete [] cEstimate;
  delete [] workSpace;

  return;
}

/*------------------------------------------------------------*/
/*----------------MTSpectrum---------------------------*/
void MTLineMon::MTSpectrum(int inorm, float dt, FSpectrum *oSpec, 
			   float *dof, float *Fvalues, fComplex *cEst)
{
	/*
	inorm=flag for choice of normalization
	dt=sampling interval (time)
	oSpec=output spectrum
	dof=degrees of freedom at each frequency
	Fvalues=Ftest value at each frequency estimate
	cEst is the estimated complex coefficient of a complex exponential (array of length paddedLength+2)
	*/

	int             kk;

	/************/
	double         *dcf, *degf;
	static int save_npoints=0;
	static int save_nWin=0;
	static float save_nwdt=0.0;
	static double *lambda=0;
	static double *tapsum=0;
	static double *tapers=0;

	// fudge
        float *ospec=new float[2*nFreq];
       
	// lambda=vector of eigenvalues
	// tapsum=sum of each taper, saved for use in adaptive weighting
	// tapers= matrix of slepian tapers packed in a 1D double array
	 
	if (chanv.mPoint!=save_npoints || nWin!=save_nWin || nwdt!=save_nwdt) {
	  // This means that we must compute tapers, etc 
	  if (lambda) delete [] lambda;
	  lambda=new double[nWin];
	  if (tapsum) delete [] tapsum;
	  tapsum=new double[nWin];
	  if (tapers) delete [] tapers;
	  tapers=new double[chanv.mPoint*nWin];
	  if (lambda==NULL || tapsum==NULL || tapers==NULL) {
	    cerr << "MTLineMon: MTSpectrum(): Insufficient memory - new returned NULL." << endl;
	    finish();
	  }
	  save_npoints=chanv.mPoint;
	  save_nWin=nWin;
	  save_nwdt=nwdt;
	  // get the set of Slepian tapers 
	  slepianTapers(lambda, tapers, tapsum);
          if (getDebug()) std::cerr << "Calculated Slepian tapers for " << save_npoints << " points with "
				    << save_nWin << " windows and NWdt = " << save_nwdt << std::endl;
	}


	int nFreq_tap=nFreq*nWin;

	// apply the taper in the loop, do this nWin times 
	float *b=new float[chanv.mPoint];
	double *amu=new double[nFreq];
	if (b==NULL || amu==NULL ) {
          cerr << "MTSpectrum(): Insufficient memory - new returned NULL." << endl;
          finish();
	}

	FSeries *f=new FSeries[nWin];
        FSpectrum **spec=new FSpectrum*[nWin];

	float *data=new float[chanv.mPoint];
 	chanv.mTS->getData(paddedLength,data);

	for (int iwin=0; iwin < nWin; iwin++) {
		kk=iwin*chanv.mPoint;

		// application of iwin-th taper   
		for (int j=0; j < chanv.mPoint; j++)  b[j]=data[j]*tapers[kk+j];
		TSeries *t=new TSeries(chanv.mTS->getStartTime(),chanv.mTS->getTStep(),chanv.mPoint,b);
                t->extend(chanv.mTS->getStartTime()+double(paddingFactor)*chanv.mTS->getInterval());   // zero padding
                f[iwin]=FSeries(*t); 
	

		// calculate the eigenspectrum 
		// mt_get_spec(b, chanv.mPoint, paddedLength,amp);		
		// double ACOnorm=sqrt(chanv.mRate*paddedLength/2.0);

		//		if (mGetFinalSpec) {
		  // get spectrum from real fourier transform    		
		//int kf=iwin*nFreq;
		//  for (int i=0;i<nFreq;i++) sqr_spec[i+kf]=padSpec[i].MagSq()/(norm*norm);
		//}


        spec[iwin]=new FSpectrum(f[iwin]);

	if (iwin==0) *oSpec=*spec[0];
	else *oSpec+=*spec[iwin];

	}
	delete [] b;
	
	double avar;
	// choice of hi-res or adaptive weighting for spectra   
	switch (mSpectralType) {
	case HIRES:

	  if (mGetFinalSpec) hiRes(spec, lambda, oSpec);
	  getFValues(f, Fvalues, tapsum,cEst,lowBin,highBin);
          for (int i=0; i < nFreq; i++) dof[i]=nWin-1;
	  break;
	  
	case ADAPTIVE:

	  avar=(*chanv.mTS)*(*chanv.mTS);

	  dcf=new double[nFreq_tap];
	  degf=new double[nFreq];
	  if (dcf==NULL || degf==NULL) {
	    cerr << "multitaper_spectrum(): Insufficient memory - new returned NULL." << endl;
	    finish();
	  }

	  adaptive(spec, dcf, lambda, amu, degf, avar);
	  getFValues(f, Fvalues, tapsum,cEst,lowBin,highBin);
	  
	  for (int i=0; i < nFreq; i++) {
	    ospec[i]=amu[i];
	    dof[i]=degf[i];
	  }

	  delete [] dcf;
	  delete [] degf;
	  break;

	default:
	  std::cerr << "# MTLineMon::multitaper_spectrum(): Do not recognize mSpectralType = " << mSpectralType << std::endl;
	  finish();
	  break;
	}

	// free up memory and return  

	delete [] data;
	delete [] spec;
	delete [] ospec;
	delete [] f;
	delete [] amu;
}




//--------------------------------------------------------
//--------------------slepianTapers----------------------
//--------------------------------------------------------
void MTLineMon::slepianTapers(double *lam, double *tapers, double *tapsum)
{
  //	get the multitaper slepian functions.
  //    lam= vector of eigenvalues
  //    tapsum=sum of each taper, saved for use in adaptive weighting
  //    tapers= matrix of slepian tapers, packed in a 1D double array

	/* check that the number of windows is not too large */
	if (nWin>=2.0*nwdt) {
	  std::cerr << "# MTLineMon::slepianTapers(): number of taper windows " << nWin << 
	    " should be LESS than 2 x NWdt = " << 2.0*nwdt << std::endl;
	  std::cerr << "Please see Percival and Walden pg 335." << std::endl;
	  finish();
	}

	double an=double(chanv.mPoint); // N in Percival and Walden
	double ww=double(nwdt)/an;	// W in Percival and Walden 

	char range = 'I'; // index range required
        char order = 'E'; // entire matrix used 
        LinAlg::CLAPACK_integer n=chanv.mPoint; // the size of the matrix
        LinAlg::CLAPACK_doublereal vl=0.0; //not used as range ='I'
        LinAlg::CLAPACK_doublereal vu=0.0; //not used as range ='I'
        LinAlg::CLAPACK_integer il=1;              // We want the first  
	LinAlg::CLAPACK_integer iu=nWin;           // nWin eigenvalues
	LinAlg::CLAPACK_doublereal abstol=LinAlg::CLAPACK_doublereal(1.0e-13); // absolute tolerance for the eigenvalues
        LinAlg::CLAPACK_doublereal* d=new LinAlg::CLAPACK_doublereal[chanv.mPoint];
        LinAlg::CLAPACK_doublereal* e=new LinAlg::CLAPACK_doublereal[chanv.mPoint-1];
	LinAlg::CLAPACK_doublereal dn=LinAlg::CLAPACK_doublereal(n);	
	double cs=cos(2.*M_PI*ww);
        for (int i=0; i < chanv.mPoint; i++) { 
	  LinAlg::CLAPACK_doublereal di=LinAlg::CLAPACK_doublereal(i);
	  d[i]=-4.0*cs*(((dn-1.)/2.-di))*(((dn-1.)/2.-di))/(dn*dn);
	  if (i !=  chanv.mPoint-1) e[i]=-4.0*(di+1.)*(dn-di-1.)/(2.*dn*dn);
  	}	
	LinAlg::CLAPACK_integer m; // number of eigenvalues found
	LinAlg::CLAPACK_integer nsplit;
	LinAlg::CLAPACK_doublereal* lambda=new LinAlg::CLAPACK_doublereal[chanv.mPoint];
	LinAlg::CLAPACK_integer* iblock=new LinAlg::CLAPACK_integer[chanv.mPoint];
	LinAlg::CLAPACK_integer* isplit=new LinAlg::CLAPACK_integer[chanv.mPoint];
	LinAlg::CLAPACK_doublereal* work=new LinAlg::CLAPACK_doublereal[4*chanv.mPoint];
	LinAlg::CLAPACK_integer* iwork=new LinAlg::CLAPACK_integer[3*chanv.mPoint];
	LinAlg::CLAPACK_integer info;

	if (d==NULL || e==NULL || lambda==NULL || iblock==NULL || isplit==NULL 
			|| work==NULL || iwork==NULL) {
          std::cerr << " MTLineMon::slepianTapers() : Insufficient memory" << std::endl;
          finish();
	}

	LinAlg::dstebz_(&range,&order,&n,&vl,&vu,&il,&iu,&abstol,d,e,&m,&nsplit,lambda,iblock,isplit,work,iwork,&info);
	if (info) {
	  std::cerr << "#MTLineMon: LinAlg::CLAPACK info returned " << info << 
	    " from dstebz_()" << std::endl;
	  finish();
	}

	delete [] e;
	delete [] iwork;
	delete [] work;

	if (getDebug()) {
	  std::cerr << "LinAlg::CLAPACK info returned " << info << std::endl;
	  std::cerr << "LinAlg::CLAPACK no of eigenvalues found  " << m << std::endl;
	  std::cerr << "LinAlg::CLAPACK no of diagonal blocks found " << nsplit << std::endl;
	  std::cerr << "eigenvalues for the eigentapers: " << std::endl;
	  for (int k=0; k < nWin; k++) std::cerr << - lambda[k] << std::endl;
	}

        // calling dstein_() to calculate eigenvectors	

        LinAlg::CLAPACK_doublereal* se=new LinAlg::CLAPACK_doublereal[chanv.mPoint];
        for (int i=0; i < chanv.mPoint; i++) { 
	  LinAlg::CLAPACK_doublereal di=LinAlg::CLAPACK_doublereal(i);
	  if (i !=  chanv.mPoint-1) se[i]=-4.0*(di+1.)*(dn-di-1.)/(2.*dn*dn);
  	}	
	LinAlg::CLAPACK_integer nEvecs=LinAlg::CLAPACK_integer(nWin);
	LinAlg::CLAPACK_doublereal* z=new LinAlg::CLAPACK_doublereal[nWin*chanv.mPoint];
	// packing defined by z(a_1,a_2) = z[(a_2)*z_dim1 + a_1]  
        // for us a1 labels the element (z_dim1=chanv.mPoint) 
        // a2 labels the eigenvalue
	LinAlg::CLAPACK_integer ldz=LinAlg::CLAPACK_integer(chanv.mPoint);
	LinAlg::CLAPACK_doublereal* swork=new LinAlg::CLAPACK_doublereal[5*chanv.mPoint];
	LinAlg::CLAPACK_integer* siwork=new LinAlg::CLAPACK_integer[chanv.mPoint];
	LinAlg::CLAPACK_integer* ifail=new LinAlg::CLAPACK_integer[nWin];

	LinAlg::dstein_(&n,d,se,&nEvecs,lambda,iblock,isplit,z,&ldz,swork,siwork,ifail,&info);

	if (info) {
	  std::cerr << "#MTLineMon: LinAlg::CLAPACK info returned " << info << 
	    " from dstein_()" << std::endl;
	  finish();
	}

	// parity convention 
	for (int k=0; k < nWin; k++) { 
	  long koffset=k*chanv.mPoint;
	  if (z[koffset+chanv.mPoint/2] < 0) {
	    for (int i=0; i<chanv.mPoint; i++) z[koffset+i]*=-1.0; 
	  }
	}

	delete [] d;
	delete [] se;
	delete [] swork;
	delete [] siwork;
	delete [] iblock;
	delete [] isplit;

	for (int k=0; k < nWin; k++) lam[k] = - double(lambda[k]);
	delete [] lambda;

	// - as we have used minus the matrix to make indexing more natural

	//******************************************************************
        //   normalize the eigentapers to preserve power for a white process
        //   i.e. they have rms value unity
        //   tapsum is the average of the eigentaper, should be near zero for
        //   antisymmetric tapers
        //*******************************************************************

	double aa;
	double tapsq;
	for (int k=0; k < nWin; k++) {
		long koffset=k*chanv.mPoint;
		tapsum[k]=0.;
		tapsq=0.;
		for (int i=0; i < chanv.mPoint; i++) {
			aa=double(z[i+koffset]);
			tapers[i+koffset]=aa;
			tapsum[k]=tapsum[k]+aa;
			tapsq=tapsq+aa*aa;
		}
		aa=sqrt(tapsq/(double) chanv.mPoint);
		tapsum[k]=tapsum[k]/aa;
		for (int i=0; i < chanv.mPoint; i++) tapers[i+koffset]/=aa;
	}


	delete [] z;

	return;
}

//--------------------------------------------------------*/
//--------------------------------------------------------*/
//----------------adaptive--------------------------------*/
int MTLineMon::adaptive(FSpectrum **spec, double *dcf, double *lambda, 
		      double *ares, double *degf, double avar)
{
  //  this version uses thomson's algorithm for calculating  the
  //  adaptive spectrum estimate Percival and Walden p 370
  const int maxIterations=50;
 
  double *sqr_spec=new double[nFreq*nWin];
  if (sqr_spec==NULL ) {
    cerr << "MTLineMon: adaptive() Insufficient memory - new returned NULL." << endl;
    finish();
  }

  for (int iwin=0; iwin < nWin; iwin++) {
    float *padSpec=new float[nFreq];
    if (padSpec==NULL ) {
      cerr << "MTLineMon: adaptive() Insufficient memory - new returned NULL." << endl;
      finish();
    }
    spec[iwin]->getData(nFreq,padSpec);
    int kf=iwin*nFreq;
    for (int i=0;i<nFreq;i++) sqr_spec[i+kf]=padSpec[i];
  }


  // set tolerance for iterative scheme exit
  double tol=3.0e-3;
  int numFailed=0;
  
  // we scale the bias by the total variance of the frequency transform
  // from zero freq to the nyquist
  // in this application we scale the eigenspectra by the bias in order to avoid
  // possible floating point overflow
  double *spw=new double[nWin];
  double *bias=new double[nWin];
  if (spw==NULL || bias==NULL) {
    std::cerr << "#MTLineMon: adaptive() insufficient memory" << std::endl;
    finish();
  }

  for (int jloop=0; jloop < nFreq; jloop++) {

    for (int iWin=0; iWin < nWin; iWin++) {    
      bias[iWin]=(1.00-lambda[iWin])*avar;
      spw[iWin]=sqr_spec[jloop+iWin*nFreq];
    }
	
    //   first guess is the average of the two
    //   lowest-order eigenspectral estimates
    ares[jloop]=(lambda[0]*spw[0]+lambda[1]*spw[1])/(lambda[0]+lambda[1]);

    //  iterative loop to find spectral weights 
    int k;
    for (k=0; k < maxIterations; k++) {
      double numer=0.00;
      double denom=0.00;

      for (int i=0; i < nWin; i++) {
	double bf=ares[jloop]/(lambda[i]*ares[jloop]+bias[i]);  //P&W Eqn(368a)
	double weight=lambda[i]*bf*bf;
	numer+=weight*spw[i];
	denom+=weight;
      }

      double newEst=numer/denom;
      if (fabs(newEst-ares[jloop]) < tol*fabs(ares[jloop])) break;
      ares[jloop]=newEst;
    }

    // flag if iteration does not converge 
    if (k >= maxIterations) numFailed++;

    // calculate degrees of freedom 
    degf[jloop]=0.0;
    for (int iWin=0; iWin < nWin; iWin++) {
      int kpoint=jloop+iWin*nFreq;
      dcf[kpoint]=ares[jloop]/(lambda[iWin]*ares[jloop]+bias[iWin]);
      degf[jloop]+=lambda[iWin]*dcf[kpoint]*dcf[kpoint];
    }
    // we normalize degrees of freedom by the weight of the first
    // eigenspectrum this way we never have fewer than two degrees of freedom
    degf[jloop]*=2./(dcf[jloop]*dcf[jloop]);
  }		

  delete [] spw;
  delete [] bias;
  delete [] sqr_spec;

  if (numFailed) std::cerr << "#MTLineMon: adaptive() " << numFailed <<  " failed iterations"  << std::endl;

  return numFailed;
}


/*--------------------------------------------------------*/
/*------------getFValues----------------------------*/
/*--------------------------------------------------------*/
void MTLineMon::getFValues(FSeries *f, float *Fvalue, double *tapsum, fComplex *cEst,int lowBin,int highBin)
{
	/*
	sri contains the real/imag part of the spectrum
	nf is the number of frequencies (2^n+1), typically

	Fvalue is the array of f-statistic values
	tapsum is fft of slepian eigentapers at zero freq
	cest is the estimated complex coefficient of a complex exponential (array of length 2 nf)
	*/

	if (lowBin<0) lowBin=0;
	if (highBin>nFreq) highBin=nFreq;

	// normalization of Slepian FFT's at DC 
	double sum=0.0;
	for (int i=0;i<nWin;i++) sum+=tapsum[i]*tapsum[i];

	// initialise coefficients 
	for (int i=0;i<nFreq;i++) {
	  cEst[i]=fComplex(0.0,0.0);
	  Fvalue[i]=0.0;
	}

	fComplex *fourier=new fComplex[nFreq];

	for (int iWin=0;iWin<nWin;iWin++) {
		double alpha=tapsum[iWin]/sum;
		f[iWin].getData(nFreq,fourier);
		for (int i=lowBin;i<highBin;i++) cEst[i]+=alpha*~fourier[i];
	}


	// store difference between estimated and actual coefficient squares 
	fComplex sumDiff;
	for (int iWin=0;iWin<nWin;iWin++) {
		double alpha=tapsum[iWin];
		f[iWin].getData(nFreq,fourier);
		for (int i=lowBin;i<nFreq;i++) {
			sumDiff=fourier[i]-alpha*~cEst[i];
			Fvalue[i]+=sumDiff.MagSq();
		}
	}


	// now finish computing the Fvalues ...
	sum*=(nWin-1);
	for (int i=lowBin;i<highBin;i++) Fvalue[i]=sum*cEst[i].MagSq()/Fvalue[i];

        // ... and finish computing the coefficients
	double ACOnorm=sqrt(chanv.mRate*paddedLength/2.0);
	for (int i=lowBin;i<highBin;i++) cEst[i]*=ACOnorm;

	delete [] fourier;
	return;
}



//--------------------------------------------------------
//-------------  hiRes  ----------------------------------
//--------------------------------------------------------
void MTLineMon::hiRes(FSpectrum **spec, double *lambda, FSpectrum *oSpec)
{
  double denominator=lambda[0];
  *spec[0]*=lambda[0];
  *oSpec=*spec[0];
  
  for (int i=1; i < nWin; i++) {
    denominator+=lambda[i];
    *spec[i]*=lambda[i];
    *oSpec+=*spec[i];
  }

  *oSpec*=(1.0/denominator);
}


void MTLineMon::readConfigFile(const char* config)
 {
   //---------------------------------  Read configuration file.
   ifstream ifs(config, ios::in);
   if (!ifs) {
     std::cerr << "# MTLineMon: unable to open configuration file " << config << std::endl;
     usage();
     exit(NO_CONFIG_FILE);
     return;
   }
 
   
   if (getDebug()) 
     std::cerr << "# MTLineMon: reading parameters from config file: " << config << std::endl;
   
   ifs >> chanv.mName >> sampleTime; 
   if (getDebug()) 
     std::cerr << "# MTLineMon: using channel " << chanv.mName << " with sample time = " 
	   << sampleTime << " seconds" <<std::endl;
   ifs.close();
}


void MTLineMon::DumpHTML(const char* channel, const char* ofile)
{
//----------------------------------  Build an html document

    html::document doc("MTLineMon statistics");
    // list<lineEntry>::iterator it;
    int row;


    doc.setBackgroundColor(html::color("wheat"));

    char cbuf[256];
    LocalStr(Now(), cbuf, "%M %d, %Y %02H:%02N");
    html::text hdr(string("MultiTaper Line Monitor Statistics at: ") + cbuf);
    hdr.setSize(html::size(2));
    html::block cb("center");
    cb.addObject(hdr);
    doc.addObject(cb);

    html::table parTbl(string("Channel:  ")+channel);
    parTbl.addColumn(" ");
    parTbl.addColumn(" ");
    row = parTbl.addRow();
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("start GPS time:  "));
    parTbl.insertData(row, 1, long(startTime));
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("current GPS time:&nbsp;&nbsp;"));
    parTbl.insertData(row, 1, long(currentTime));
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("live time, sec:  "));
    parTbl.insertData(row, 1, long(liveTime));

    doc.addObject(parTbl);


//----------------------------------  Add line table.

    html::table lineTbl("Detected Lines");

    lineTbl.addColumn("frequency");
    lineTbl.addColumn("&nbsp;&nbsp;&nbsp;amplitude");
    lineTbl.addColumn("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;F-value");
    lineTbl.addColumn("&nbsp;&nbsp;&nbsp;&nbsp;entries");
    lineTbl.addColumn("&nbsp;&nbsp;duty cycle");

    row = lineTbl.addRow();
    lineTbl.insertData(row, 0, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Hz");
    lineTbl.insertData(row, 4, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%");

    for (int i=0 ; i < nRemoved ; i++) {
      double duty = 0.;

      row = lineTbl.addRow();
      int nn = int(100.*lineList[i].index*0.5*chanv.mRate/nFreq);
//  	if(!(nn-(nn/10)*10)) nn++;
      double freq = double(nn)/100.;
      lineTbl.insertData(row, 0, freq);
      sprintf(cbuf,"&nbsp;&nbsp;&nbsp;&nbsp;%13.2f",(lineList[i].FC).Mag());
      lineTbl.insertData(row, 1, cbuf);
      sprintf(cbuf,"&nbsp;&nbsp;&nbsp;%9.2f",lineList[i].fvalue);
      lineTbl.insertData(row, 2, cbuf);
      lineTbl.insertData(row, 3, long(lineCount[lineList[i].index]));
      duty=(100.*lineCount[lineList[i].index])/(mSegment+1);
      sprintf(cbuf,"&nbsp;&nbsp;&nbsp;%9.2f",duty);
      lineTbl.insertData(row, 4, cbuf);
    }

    doc.addObject(lineTbl);

    html::table legTbl("Legend");

    legTbl.addColumn("");
    row = legTbl.addRow();

    legTbl.insertData(row, 0, 
     "Amplitude - line amplitude");
    row = legTbl.addRow();
    legTbl.insertData(row, 0, 
     "F-value - statistical measure of significance of line");
    sprintf(cbuf," (critical value %9.2f)",fCritical);
    legTbl.insertData(row, 0, cbuf);


    row = legTbl.addRow();
    legTbl.insertData(row, 0, 
     "duty cycle - fraction of time when F-value for is above the threshold");

    doc.addObject(legTbl);

//----------------------------------  Write the document

    ofstream f(string(ofile).c_str());
    html::writer w(f);
    doc.write(w);
}




void MTLineMon::usage(void)
{
  std::cerr << "Command syntax is:" << std::endl;
  std::cerr << "\t MTLineMon -conf <config> [-adaptive or -hires] "
	    << "[-nWin <number of Windows: default 5>] [-nwdt <NWdt: default 3.0>] "
	    << "[-paddingfactor <default: 4>] "
	    << "[-freqRange <default: 0 16384>] "
	    << "[-debug] "
	    << "[-nofinalspec] "
	    << "[-log] "
	    << "[-maxSegments <default -1 (no limit)>] " 
	    << "[-help] this message" << std::endl;
  finish();
}


